"""
Dimensionality measures
"""
import numpy as np
from sklearn.decomposition import PCA

def t2(X, y):
    """
    Calculates the Average number of features per points (T2) metric. 

    To obtaint this measure, the number of dimensions describing the dataset is divided by the number of instances.

    .. math::

        T2=\\frac{m}{n}

    :type X: array-like, shape (n_samples, n_features)
    :param X: Dataset
    :type y: array-like, shape (n_samples)
    :param y: Labels of binary classification task ([0,1])

    :rtype: float
    :returns: T2 score
    """
    X = np.copy(X)
    return X.shape[1]/X.shape[0]

def t3(X, y):
    """
    Calculates the Average number of PCA dimensions per points (T3) metric.

    To obtain this measure, first, the number of PCA components needed to represent 95% of data variability is calculated. Then, the value is divided by the instance number in the dataset.

    .. math::

        T3=\\frac{m'}{n}

    :type X: array-like, shape (n_samples, n_features)
    :param X: Dataset
    :type y: array-like, shape (n_samples)
    :param y: Labels of binary classification task ([0,1])

    :rtype: float
    :returns: T3 score
    """
    X = np.copy(X)
    y = np.copy(y)
    
    pca = PCA(n_components=0.95)
    X_new = pca.fit_transform(X,y)

    return X_new.shape[1]/X.shape[0]

def t4(X, y):
    """
    Calculates the Ration of the PCA dimension to the original dimension (T4) metric.

    To obtain this measure, the number of PCA components needed to represent 95% of data variability is divided by the original number of dimensions. This measure describes the proportion of relevant dimensions in the dataset.

    .. math::

        T4=\\frac{m'}{m}

    :type X: array-like, shape (n_samples, n_features)
    :param X: Dataset
    :type y: array-like, shape (n_samples)
    :param y: Labels of binary classification task ([0,1])

    :rtype: float
    :returns: T4 score
    """
    X = np.copy(X)
    y = np.copy(y)
    
    pca = PCA(n_components=0.95)
    X_new = pca.fit_transform(X,y)

    return X_new.shape[1]/X.shape[1]


"""
Linearity measures for regression task
"""
import numpy as np
from sklearn.linear_model import LinearRegression

def l1(X, y, normalize=True):
    """
    Calculates the mean absolute error (L1) metric. 

    Measure returns average error of linear regression model. By default performs a 0-1 interval normalization.

    .. math::

        L1=\sum_{i=1}^{n}\\frac{|\epsilon_i|}{n}

    :type X: array-like, shape (n_samples, n_features)
    :param X: Dataset
    :type y: array-like, shape (n_samples)
    :param y: Labels for regression task

    :rtype: float
    :returns: L1 score
    """
    if normalize:
        X = np.copy(X)
        y = np.copy(y)
        for f_id in range(X.shape[1]):
            X[:,f_id] -= np.min(X[:,f_id])
            X[:,f_id] /= np.max(X[:,f_id])
        y -= np.min(y)
        y /=np.max(y)

    linreg = LinearRegression().fit(X,y)
    pred = linreg.predict(X)
    residuals = np.abs(y-pred)

    return np.sum(residuals)/X.shape[0]

def l2(X, y, normalize = True):
    """
    Calculates the residuals variance (L2) metric. 

    Measure returns average of squared residuals of linear regression model. By default performs a 0-1 interval normalization.

    .. math::

        L2=\sum_{i=1}^{n}\\frac{\epsilon_i^2}{n}

    :type X: array-like, shape (n_samples, n_features)
    :param X: Dataset
    :type y: array-like, shape (n_samples)
    :param y: Labels for regression task

    :rtype: float
    :returns: L2 score
    """
    if normalize:
        X = np.copy(X)
        y = np.copy(y)
        for f_id in range(X.shape[1]):
            X[:,f_id] -= np.min(X[:,f_id])
            X[:,f_id] /= np.max(X[:,f_id])
        y -= np.min(y)
        y /=np.max(y)

    linreg = LinearRegression().fit(X,y)
    pred = linreg.predict(X)
    residuals = y-pred

    return np.sum(np.power(residuals,2))/X.shape[0]


"""Regression tests."""
import problexity as px
from sklearn.datasets import make_regression
import numpy as np

np.random.seed(123)
    
def _get_comparison(metric, args=[]):
    reps = 10
    res = np.zeros((reps, 2))
    rs = np.random.randint(100,1000,reps)

    for r in range(reps):
        X_simple, y_simple = make_regression(n_samples=300, n_features=50,
                        n_informative=50, noise=0, random_state=rs[r])

        X_complex, y_complex = make_regression(n_samples=300, n_features=100,
                        n_informative=10, noise=1200, random_state=rs[r])

        r_s = metric(X_simple, y_simple, *args)
        r_c = metric(X_complex, y_complex, *args)

        res[r, 0] = r_s
        res[r, 1] = r_c

    compare = np.mean(res[:,0]<res[:,1])
    return compare
 
### C1 tests ###

def test_C1():
    metric = px.regression.c1
    assert _get_comparison(metric)<0.5

### C2 tests ###

def test_C2():
    metric = px.regression.c2
    assert _get_comparison(metric)<0.5

## C3 tests ###

def test_C3():
    metric = px.regression.c3
    assert _get_comparison(metric)>0.5

def test_C3_l():
    metric = px.regression.c3
    assert _get_comparison(metric, [False])>0.5

## C4 tests ###

def test_C4():
    metric = px.regression.c4
    assert _get_comparison(metric)>0.5

## L1 tests ###

def test_L1():
    metric = px.regression.l1
    assert _get_comparison(metric)>0.5

## L2 tests ###

def test_L2():
    metric = px.regression.l2
    assert _get_comparison(metric)>0.5

### S1 tests ###

def test_S1():
    metric = px.regression.s1
    assert _get_comparison(metric)>0.5

### S2 tests ###

def test_S2():
    metric = px.regression.s2
    assert _get_comparison(metric)>0.5

### S3 tests ###

def test_S3():
    metric = px.regression.s3
    assert _get_comparison(metric)>0.5

### L3 tests ###

def test_L3():
    metric = px.regression.l3
    assert _get_comparison(metric)>0.5

### S4 tests ###

def test_S4():
    metric = px.regression.s4
    assert _get_comparison(metric)>0.5

### T2 tests ###

def test_T2():
    metric = px.regression.t2
    assert _get_comparison(metric)<0.5

### Test ComplexityCalculator ###
def test_ComplexityCalculator():
    c = px.ComplexityCalculator(mode='regression')
    reps = 5
    rs = np.random.randint(100,1000,reps)

    for r in range(reps):
        X_simple, y_simple = make_regression(n_samples=50, n_features=50,
                        n_informative=50, noise=0, random_state=rs[r])

        X_complex, y_complex = make_regression(n_samples=50, n_features=100,
                        n_informative=10, noise=1200, random_state=rs[r])

        c.fit(X_simple, y_simple).score()
        c.fit(X_complex, y_complex).score()
        
def test_ComplexityCalculator_report():
    c = px.ComplexityCalculator(mode = 'regression')

    X_simple, y_simple = make_regression(n_samples=50, n_features=50,
                        n_informative=50, noise=0, random_state = 123)

    report = c.fit(X_simple, y_simple).report()
    assert isinstance(report, dict)



"""
Feature correlation measures
"""
from scipy.stats import spearmanr
import numpy as np
from sklearn.linear_model import LinearRegression

def c1(X, y, normalize=True):
    """
    Calculates the maximum feature correlationto the output (C1) metric. 

    Measure returns maximum value out of all feature-output Spearman correlation absolute value. Higher values indicate simpler problems.

    .. math::

        C1=max_{j=1,..,d}|\\rho(x^j, y)|

    :type X: array-like, shape (n_samples, n_features)
    :param X: Dataset
    :type y: array-like, shape (n_samples)
    :param y: Labels for regression task

    :rtype: float
    :returns: C1 score
    """
    X = np.copy(X)
    y = np.copy(y)

    if normalize:
        for f_id in range(X.shape[1]):
            X[:,f_id] -=np.min(X[:,f_id])
            X[:,f_id] /=np.max(X[:,f_id])
        y -= np.min(y)
        y /=np.max(y)

    corr = np.zeros((X.shape[1]))
    for f_id in range(X.shape[1]):
        corr[f_id] = np.abs(spearmanr(X[:,f_id], y).correlation)
    
    return np.max(corr)

def c2(X, y, normalize = True):
    """
    Calculates the average feature correlationto the output (C2) metric. 

    Measure returns average value of all feature-output Spearman correlation absolute value. Higher values indicate simpler problems.

    .. math::

        C2=\sum^{d}_{j=1}\\frac{|\\rho(x^j, y)|}{d}

    :type X: array-like, shape (n_samples, n_features)
    :param X: Dataset
    :type y: array-like, shape (n_samples)
    :param y: Labels for regression task

    :rtype: float
    :returns: C2 score
    """
    X = np.copy(X)
    y = np.copy(y)

    if normalize:
        for f_id in range(X.shape[1]):
            X[:,f_id] -=np.min(X[:,f_id])
            X[:,f_id] /=np.max(X[:,f_id])
        y -= np.min(y)
        y /=np.max(y)

    corr = np.zeros((X.shape[1]))
    for f_id in range(X.shape[1]):
        corr[f_id] = np.abs(spearmanr(X[:,f_id], y).correlation)
    
    return np.sum(corr)/X.shape[1]    

def _c3_l(_X, _y):
    high_correlation_treshold = 0.9

    njn = np.zeros((_X.shape[1]))

    for f_id in range(_X.shape[1]):
        # For each feature a linear regression model is fitted on data.
        linreg = LinearRegression().fit(_X[:,f_id].reshape(-1,1),_y)
        y_pred = linreg.predict(_X[:,f_id].reshape(-1,1))

        # The residual values are calculated based on predictions.
        residuals = np.abs(_y - y_pred)
        mask = np.ones_like(residuals).astype(bool)

        # The initial correlation for feature is calculated.
        f_correlation = np.abs(spearmanr(_X[:,f_id], _y).correlation)
        while(f_correlation<=high_correlation_treshold):
            # The sample with a largest residual value is selected to be removed, and its residual value is substituted with -infinity.
            to_remove = np.argmax(residuals)
            residuals[to_remove]=-np.inf
            mask[to_remove]=0

            # The new correlation value is calculated. The loop is repeated until a high correlation value is achieved.
            f_correlation = np.abs(spearmanr(_X[:,f_id][mask], _y[mask]).correlation)
           
        num_removed = np.sum(mask==0)
        njn[f_id] = num_removed/_X.shape[0]
    
    return np.min(njn)    

def _c3_h(_X, _y):
    high_correlation_treshold = 0.9

    njn = np.zeros((_X.shape[1]))

    n_samples = _X.shape[0]
    
    for f_id in range(_X.shape[1]):
        # For each feature a linear regression model is fitted on data.
        linreg = LinearRegression().fit(_X[:,f_id].reshape(-1,1),_y)
        y_pred = linreg.predict(_X[:,f_id].reshape(-1,1))

        # The residual values are calculated based on predictions.
        residuals = np.abs(_y - y_pred)
        
        sorter = np.argsort(-residuals)
        
        # Objects are sorted based on residual value -- starting from the highest.
        rX = np.copy(_X[sorter])
        ry = np.copy(_y[sorter])

        # Initialize the step and split point.
        step = n_samples / 2
        head = 0
        
        while (True):
            # Current X and y, based on current split point.
            __X = rX[int(head):, f_id]
            __y = ry[int(head):]
            
            corr = np.abs(spearmanr(__X, __y).correlation)
            if np.isnan(corr):
                corr = 1
            
            if corr > high_correlation_treshold:
                # Step in the right direction -- lower the step.
                step /= 2
                head -= step
            else:
                head += step
                
            # If step is smaller than 1/2 instance -- break the loop.
            if step < .5:
                break

        # Establish number or removed objects based on split position.
        num_removed = int(np.rint(head))
        njn[f_id] = num_removed/_X.shape[0]
        
    return np.min(njn)

def c3(X, y, is_optimized=True, normalize=True):
    """
    Calculates the individual feature efficiency (C3) metric. 

    Measure is calculated based on a number of examples that have to be removed in order to obtain a high correlation value. Removes samples based on residual value of linear regression model. The is_optimized flag value allows using optimized algorithm, based on divide and conquer strategy.

    .. math::

        C3=min_{j=1}^{d}\\frac{n^j}{n}

    :type X: array-like, shape (n_samples, n_features)
    :param X: Dataset
    :type y: array-like, shape (n_samples)
    :param y: Labels for regression task

    :rtype: float
    :returns: C3 score
    """
    X = np.copy(X)
    y = np.copy(y)

    if normalize:
        for f_id in range(X.shape[1]):
            X[:,f_id] -=np.min(X[:,f_id])
            X[:,f_id] /=np.max(X[:,f_id])
        y -= np.min(y)
        y /=np.max(y)

    return _c3_h(X,y) if is_optimized else _c3_l(X,y)

def c4(X, y, normalize = True):
    """
    Calculates the collective feature efficiency (C4) metric. 

    It sequentially analyzes the features with the greatest correlation to the output until all the features are used or all instances are removed. Samples with low resudual value are removed. A metric is computed based on the number of samples remaining after removal procedure. By default, 0-1 interval normalization is used. The iterations limit of 1000 was introduced.

    .. math::

        C4=\\frac{\#\{x_i||\epsilon_i|>0.1\}_{T_l}}{n}

    :type X: array-like, shape (n_samples, n_features)
    :param X: Dataset
    :type y: array-like, shape (n_samples)
    :param y: Labels for regression task

    :rtype: float
    :returns: C4 score
    """
    treshold = 0.1
    features_used = np.zeros((X.shape[1])).astype(bool)
    corr = np.zeros((X.shape[1]))

    _X = np.copy(X)
    _y = np.copy(y)

    if normalize:
        for f_id in range(X.shape[1]):
            _X[:,f_id] -=np.min(_X[:,f_id])
            _X[:,f_id] /=np.max(_X[:,f_id])
        _y -= np.min(_y)
        _y /=np.max(_y)

    while (np.sum(features_used)<X.shape[1]) and len(_y)>0:

        for f_id in range(X.shape[1]):
            corr[f_id] = np.abs(spearmanr(_X[:,f_id], _y).correlation)
        corr[features_used]=-np.inf

        curr_f = np.argmax(corr)
        features_used[curr_f] = 1

        linreg = LinearRegression().fit(_X[:,curr_f].reshape(-1,1),_y)
        y_pred = linreg.predict(_X[:,curr_f].reshape(-1,1))

        residuals = np.abs(_y - y_pred)

        small_residuals_mask = np.zeros_like(residuals)
        small_residuals_mask[residuals<=treshold] = 1

        remaining_n = np.sum(small_residuals_mask==False)

        _X = _X[small_residuals_mask==False]
        _y = _y[small_residuals_mask==False]

    return remaining_n/X.shape[0]


from .feature_based import f1, f1v, f2, f3, f4
from .linearity import l1, l2, l3
from .neighborhood import n1, n2, n3, n4, t1, lsc
from .network import density, clsCoef, hubs
from .dimensionality import  t2, t3, t4
from .class_imbalance import c1, c2

__all__ = [
    "f1", 
    "f1v", 
    "f2",
    "f3",
    "f4",
    "l1",
    "l2",
    "l3",
    "n1",
    "n2",
    "n3",
    "n4",
    "t1",
    "lsc",
    "density",
    "clsCoef",
    "hubs",
    "t2",
    "t3",
    "t4",
    "c1",
    "c2"
]



import numpy as np
from scipy.spatial import distance_matrix
from .gower import gower_matrix
from igraph import Graph

def _get_graph(X, y):
    X = np.copy(X)
    y = np.copy(y)

    epsilon= 0.15
    dist = gower_matrix(X)

    for i in range(y.shape[0]):
        dist[i,i]=np.nan
    
    #normalize
    dist = dist-np.nanmin(dist)
    dist = dist/np.nanmax(dist)

    # build graph
    edges = np.zeros(dist.shape)
    edges[dist<epsilon] = 1

    # rm
    dist_y = distance_matrix(y[:, np.newaxis], y[:, np.newaxis])
    other_classes = np.argwhere(dist_y==1)
    edges[other_classes[:,0], other_classes[:,1]] = 0

    return edges


def density(X, y):
    """
    Calculates the Density metric.

    Generates an epsilon-Nearest Neighbours graph. The epsilon value is set to 0.15. The edges are selected based on the Gower distance between samples, normalized to the range between 0 and 1. Edges between instances of distinct classes are removed. The measure calculates the number of edges in the final graph divided by the total possible number of edges.
    
    .. math::

        Density =1 - \\frac{2|E|}{n(n-1)}

    :type X: array-like, shape (n_samples, n_features)
    :param X: Dataset
    :type y: array-like, shape (n_samples)
    :param y: Labels of binary classification task ([0,1])

    :rtype: float
    :returns: Density score
    """
    
    X = np.copy(X)
    y = np.copy(y)
    
    edges = _get_graph(X, y)

    return 1-(np.sum(edges)/(y.shape[0]*(y.shape[0]-1)))


def clsCoef(X, y):
    """
    Calculates the Clustering Coefficient metric.

    Generates an epsilon-Nearest Neighbours graph. The epsilon value is set to 0.15. The edges are selected based on the Gower distance between samples, normalized to the range between 0 and 1. Edges between instances of distinct classes are removed. The neighborhood of each vertex is calculated – the instances directly connected to it. Then, the number of edges between the sample's neighbors is calculated and divided by the maximum possible number of edges between them. The final measure is calculated based on the neighborhood of each point.
    
    .. math::

        ClsCoef=1-\\frac{1}{n}\sum^{n}_{i=1}\\frac{2|e_{jk} : v_j, v_k \in N_i|}{k_i(k_i-1)}

    :type X: array-like, shape (n_samples, n_features)
    :param X: Dataset
    :type y: array-like, shape (n_samples)
    :param y: Labels of binary classification task ([0,1])

    :rtype: float
    :returns: Clustering Coefficient score
    """
    
    X = np.copy(X)
    y = np.copy(y)
    
    edges = _get_graph(X, y)

    sum_neighbours = np.sum(edges, axis=0)
    sum_conncetions = np.zeros(y.shape[0])

    #edges between its neighbours
    for i, item_egdes in enumerate(edges):
        neighbours = np.argwhere(item_egdes==1).flatten()

        #connections between n and other neighbours
        sum_conncetions[i] = np.sum([edges[n, neighbours] for n in neighbours])

    mask = (sum_neighbours-1)>0

    sum_neighbours = sum_neighbours[mask]
    sum_conncetions = sum_conncetions[mask]

    return 1 - (np.sum(sum_conncetions/(sum_neighbours*(sum_neighbours-1))))/y.shape[0]

def hubs(X, y):
    """
    Calculates the Hubs metric.

    Generates an epsilon-Nearest Neighbours graph. The epsilon value is set to 0.15. The edges are selected based on the Gower distance between samples, normalized to the range between 0 and 1. Edges between instances of distinct classes are removed. The neighborhood of each vertex is obtained – the instances directly connected to it. The measure scores each sample by the number of connections to neighbors, weighted by the number of connections the neighbors have.

    .. math::

        Hubs=1-\\frac{1}{n}\sum^{n}_{i=1}hub(v_i)

    :type X: array-like, shape (n_samples, n_features)
    :param X: Dataset
    :type y: array-like, shape (n_samples)
    :param y: Labels of binary classification task ([0,1])

    :rtype: float
    :returns: Hubs score
    """
    
    X = np.copy(X)
    y = np.copy(y)
    
    edges = _get_graph(X, y)

    g = Graph.Adjacency(edges)
    hub = g.hub_score()
    return 1 - (np.sum(hub))/y.shape[0]


"""
Neighborhood measures
"""
import numpy as np
from scipy.spatial import distance_matrix
from scipy.sparse import csr_matrix
from scipy.sparse.csgraph import minimum_spanning_tree
from sklearn.neighbors import KNeighborsClassifier, kneighbors_graph

def n1(X, y):
    """
    Calculates the Fraction of borderline points (N1) metric.

    The Minimum Spanning Three is generated over input instances. The measure is computed by calculating the number of edges in the MST between instances of different classes over a total number of samples. 

    .. math::

        N1=\\frac{1}{n} \sum^{n}_{i=1}I((x_i, x_j) \in MST \wedge y_i \\neq y_j)

    :type X: array-like, shape (n_samples, n_features)
    :param X: Dataset
    :type y: array-like, shape (n_samples)
    :param y: Labels of binary classification task ([0,1])

    :rtype: float
    :returns: N1 score
    """

    X = np.copy(X)
    y = np.copy(y)

    dist = distance_matrix(X, X)
    graph = csr_matrix(dist)
    mst = minimum_spanning_tree(graph)
    mst = mst.toarray()
    coordinates = np.argwhere(mst>0)

    return (np.sum(np.sum(y[coordinates], axis=1)==1)/2)/y.shape[0]

def n2(X, y):
    """
    Calculates the Ratio of intra/extra class NN distance (N2) metric.

    The measure depends on the distances of each problem instance to its nearest neighbor of the same class and the distance to the nearest neighbor of a different class. According to the proportions of those values, the final measure is calculated.

    .. math::

        N2=\\frac{intra\_extra}{1+intra\_extra}

    :type X: array-like, shape (n_samples, n_features)
    :param X: Dataset
    :type y: array-like, shape (n_samples)
    :param y: Labels of binary classification task ([0,1])

    :rtype: float
    :returns: N2 score
    """

    X = np.copy(X)
    y = np.copy(y)

    X_0 = X[y==0]
    X_1 = X[y==1]

    try:
        dm_intra_0 = distance_matrix(X_0, X_0)
        dm_intra_1 = distance_matrix(X_1, X_1)
        intra_0 = dm_intra_0[np.arange(len(X_0)),np.argsort(dm_intra_0)[:,1]]
        intra_1 = dm_intra_1[np.arange(len(X_1)),np.argsort(dm_intra_1)[:,1]]

        dm_extra_0 = distance_matrix(X_0, X_1)
        dm_extra_1 = distance_matrix(X_1, X_0)
        extra_0 = dm_extra_0[np.arange(len(X_0)),np.argsort(dm_extra_0)[:,1]]
        extra_1 = dm_extra_1[np.arange(len(X_1)),np.argsort(dm_extra_1)[:,1]]
    except:
        return np.nan

    intra_extra = (np.sum(intra_0)+np.sum(intra_1))/(np.sum(extra_0)+np.sum(extra_1))
    return intra_extra/(1+intra_extra)


def n3(X, y):
    """
    Calculates the Error rate of NN classifier (N3) metric.

    Measure is determined by the error rate of the One Nearest Neighbor Classifier in the Leave One Out evaluation protocol.

    .. math::

        N3=\\frac{\sum^{n}_{i=1}I(NN(x_i) \\neq y_i)}{n}

    :type X: array-like, shape (n_samples, n_features)
    :param X: Dataset
    :type y: array-like, shape (n_samples)
    :param y: Labels of binary classification task ([0,1])

    :rtype: float
    :returns: N3 score
    """
    X = np.copy(X)
    y = np.copy(y)

    neighbors = kneighbors_graph(X, n_neighbors=1, mode='connectivity', include_self=False).nonzero()[1]

    correct = y == y[neighbors]
    acc = np.sum(correct)/len(correct)

    return 1-acc

def n4(X, y):
    """
    Calculates the Nonlinearity of NN classifier (N4) metric.

    The measure is determined by the error rate of k - Nearest Neighbor Classifier on synthetic points, generated by linearly interpolating original instances. The Classifier is fitted on original points and evaluated on synthetic instances.

    .. math::

        N4=\\frac{1}{l}\sum^{l}_{i=1}I(NN_T(x'_i) \\neq y'_i)

    :type X: array-like, shape (n_samples, n_features)
    :param X: Dataset
    :type y: array-like, shape (n_samples)
    :param y: Labels of binary classification task ([0,1])

    :rtype: float
    :returns: N4 score
    """

    X = np.copy(X)
    y = np.copy(y)

    try:
        pairs_0 = np.array([np.random.choice(np.argwhere(y==0).flatten(), 2, replace=False) for i in range(np.sum(y==0))])
        pairs_1 = np.array([np.random.choice(np.argwhere(y==1).flatten(), 2, replace=False) for i in range(np.sum(y==1))])
    except:
        # Protect againt single class sample
        pairs_0 = np.array([np.random.choice(np.argwhere(y==0).flatten(), 2, replace=True) for i in range(np.sum(y==0))])
        pairs_1 = np.array([np.random.choice(np.argwhere(y==1).flatten(), 2, replace=True) for i in range(np.sum(y==1))])

    pairs = np.concatenate((pairs_0, pairs_1), axis=0)
    y_new = np.concatenate((np.zeros(pairs_0.shape[0]), np.ones(pairs_1.shape[0])), axis=0)
    rand = np.random.random(pairs.shape[0]).reshape(-1,1)

    distance = X[pairs[:,0]]-X[pairs[:,1]]
    rand_distance = rand * distance

    X_new = X[pairs[:,1]] + rand_distance

    knn = KNeighborsClassifier()
    pred = knn.fit(X, y).predict(X_new)

    return np.sum(pred!=y_new)/y_new.shape[0]

def _find_radius(D, i):
    j = np.argmin(D[i])
    di = D[i, j]
    k = np.argmin(D[j])
    if i == k:
        return di/2
    else:
        dt = _find_radius(D, j)
        return di - dt

def t1(X, y):
    """
    Calculates the Fraction of hyperspheres covering data (T1) metric.

    The measure is described by the number of hyperspheres needed to cover the data divided by a number of instances. First, a hypersphere is generated for each problem sample. A sample lies in the center of the hypersphere. Its radius is dependent on the distance to the instance of another class. The hyperspheres are eliminated if a different hypersphere already covers the center instance. The elimination starts from the hyperspheres with the largest radiuses and continues to the ones with smaller radiuses. The hyperspheres that were not eliminated are taken into account during the calculation of complexity.
    
    .. math::

        T1=\\frac{\#Hyperspheres(T)}{n}

    :type X: array-like, shape (n_samples, n_features)
    :param X: Dataset
    :type y: array-like, shape (n_samples)
    :param y: Labels of binary classification task ([0,1])

    :rtype: float
    :returns: T1 score
    """

    X = np.copy(X)
    y = np.copy(y)

    dist = distance_matrix(X, X)
    dist_y = distance_matrix(y[:, np.newaxis], y[:, np.newaxis])

    same_classes = np.argwhere(dist_y==0)
    dist[same_classes[:,0], same_classes[:,1]] = np.inf

    radiuses = np.zeros((y.shape[0]))

    for x_id in range(y.shape[0]):
        radiuses[x_id] = _find_radius(dist, x_id)

    #start with largest radiuses
    ordered_r = np.flip(np.argsort(radiuses))
    covered = np.zeros((X.shape[0]))

    hyper = 0
    for r_id in ordered_r:

        class_y = y[r_id]
        radius = radiuses[r_id]
        point = X[r_id]

        if np.isnan(radius): #already absorbed
            continue

        dist_to_point = distance_matrix(X, point[np.newaxis, :])
        dist_to_point[y!=class_y] = np.inf

        dist_to_point = dist_to_point.flatten()
        absorbed = np.argwhere(dist_to_point<radius)

        hyper+=1

        covered[absorbed] = True
        covered[r_id] = True

        radiuses[absorbed] = np.nan

        if False not in covered:
            break

    return hyper/X.shape[0]

def lsc(X, y):
    """
    Calculates the Local set average cardinality (LSC) metric.

    The measure is dependent on the distances between instances and the distances to the instances' nearest enemies – the nearest sample of the opposite class. The number of cases that lie closer to the sample than its closest enemy is taken into account during calculation. 
    
    .. math::

        LSC=1-\\frac{1}{n_2}\sum^{n}_{i=1} |LS(x_i)|

    :type X: array-like, shape (n_samples, n_features)
    :param X: Dataset
    :type y: array-like, shape (n_samples)
    :param y: Labels of binary classification task ([0,1])

    :rtype: float
    :returns: LSC score
    """

    X = np.copy(X)
    y = np.copy(y)

    dist = distance_matrix(X, X)
    dist_y = distance_matrix(y[:, np.newaxis], y[:, np.newaxis])

    same_classes = np.argwhere(dist_y==0)
    dist_enemies = np.copy(dist)
    dist_enemies[same_classes[:,0], same_classes[:,1]] = np.inf

    nearest_enemies = np.argmin(dist_enemies, axis=0)
    ne_dist = dist[np.arange(y.shape[0]), nearest_enemies]

    ls = np.argwhere(dist<ne_dist)
    return 1 - (ls.shape[0]/np.power(y.shape[0],2))


"""
Class imbalance measures
"""

import numpy as np

def c1(X, y):
    """
    Calculates the Entropy of Class Proportions (C1) metric.

    .. math::

        C1=1+\\frac{1}{log(n_c)}\sum^{n_c}_{i=1}p_{c_{i}}log(p_{c_{i}})

    :type X: array-like, shape (n_samples, n_features)
    :param X: Dataset
    :type y: array-like, shape (n_samples)
    :param y: Labels of binary classification task ([0,1])

    :rtype: float
    :returns: C1 score
    """
    y = np.copy(y)

    classes = np.unique(y).shape[0]
    p = np.zeros((classes))
    for c in range(classes):
        p[c] = np.sum(y==c)/y.shape[0]

    return 1+(1/np.log(classes))*np.sum(p*np.log(p))

def c2(X, y):
    """
    Calculates the Imbalance Ratio (C2) metric.

    .. math::

        C2=1-\\frac{1}{IR}

    :type X: array-like, shape (n_samples, n_features)
    :param X: Dataset
    :type y: array-like, shape (n_samples)
    :param y: Labels of binary classification task ([0,1])

    :rtype: float
    :returns: C2 score
    """
    y = np.copy(y)

    classes = np.unique(y).shape[0]
    p = np.zeros((classes))
    for c in range(classes):
        p[c] = np.sum(y==c)/np.sum(y!=c)

    ir = 0.5 * np.sum(p)
    return 1 - (1/ir)


"""
Feature-based measures
"""
import numpy as np

def f1(X, y):
    """
    Calculates the Maximum Fisher's discriminant ratio (F1) metric.

    Measure describes the overlap of feature values in each class.

    .. math::

        F1=\\frac{1}{1+max^{m}_{i=1}r_{f_{i}}}

    :type X: array-like, shape (n_samples, n_features)
    :param X: Dataset
    :type y: array-like, shape (n_samples)
    :param y: Labels of binary classification task ([0,1])

    :rtype: float
    :returns: F1 score
    """
    
    X = np.copy(X)
    y = np.copy(y)

    X_0 = X[y==0]
    X_1 = X[y==1]

    try:
        X_0_prop = X_0.shape[0]/X.shape[0]
        X_1_prop = X_1.shape[0]/X.shape[0]
    except:
        return np.nan

    X_0_mean = np.mean(X_0, axis = 0)
    X_1_mean = np.mean(X_1, axis = 0)

    X_0_std = np.std(X_0, axis = 0)
    X_1_std = np.std(X_1, axis = 0)

    l = (X_0_prop*X_1_prop*np.power((X_0_mean - X_1_mean),2)) + (X_1_prop*X_0_prop*np.power((X_1_mean - X_0_mean),2))
    m = (X_0_prop*(np.power(X_0_std,2))) + (X_1_prop*(np.power(X_1_std,2)))
    r_all = l/m
    
    return 1 / (1+np.nanmax(r_all))

def f1v(X, y):
    """
    Calculates the Directional vector maximum Fisher's discriminant ratio (F1v) metric.

    .. math::

        F1v=\\frac{1}{1+dF}

    :type X: array-like, shape (n_samples, n_features)
    :param X: Dataset
    :type y: array-like, shape (n_samples)
    :param y: Labels of binary classification task ([0,1])

    :rtype: float
    :returns: F1v score
    """
    
    X = np.copy(X)
    y = np.copy(y)

    X_0 = X[y==0]
    X_1 = X[y==1]

    X_0_prop = X_0.shape[0]/X.shape[0]
    X_1_prop = X_1.shape[0]/X.shape[0]

    X_0_mean = np.mean(X_0, axis = 0)
    X_1_mean = np.mean(X_1, axis = 0)

    sigm0 = np.zeros((X.shape[1],X.shape[1]))
    sigm1 = np.zeros((X.shape[1],X.shape[1]))
    
    for i in range(X.shape[1]):
        for j in range(X.shape[1]):
            sigm0[i,j] = np.mean((X_0[:,i] - X_0_mean[i]) * (X_0[:,j] - X_0_mean[j]))
    
    for i in range(X.shape[1]):
        for j in range(X.shape[1]):
            sigm1[i,j] = np.mean((X_1[:,i] - X_1_mean[i]) * (X_1[:,j] - X_1_mean[j]))
    
    B = np.dot((X_0_mean - X_1_mean).reshape(-1,1),(X_0_mean - X_1_mean).reshape(1,-1))

    W = sigm0*X_0_prop + sigm1*X_1_prop
    try:
        W_1 = np.linalg.pinv(W)
    except:
        # Will return nan in case only one class instances occur -- numpy throws: numpy.linalg.LinAlgError: SVD did not converge
        return np.nan

    d = np.dot(W_1,(X_0_mean - X_1_mean))
    dt = d.T

    df = (np.dot(np.dot(dt,B),d))/(np.dot(np.dot(dt,W),d))
    if np.isnan(df):
        df = 0

    return 1/(1+df)


def f2(X, y):
    """
    Calculates the Volume of overlapping region (F2) metric.

    Describes the overlap of the feature values within the classes. The measure is determined by the minimum and maximum values of features in the class. The overlap is then calculated and normalized by the range of values in each class.

    .. math::

        F2=\prod^{m}_{i}{\\frac{overlap(f_i)}{range(f_i)}}

    :type X: array-like, shape (n_samples, n_features)
    :param X: Dataset
    :type y: array-like, shape (n_samples)
    :param y: Labels of binary classification task ([0,1])

    :rtype: float
    :returns: F2 score
    """

    X = np.copy(X)
    y = np.copy(y)

    X_0 = X[y==0]
    X_1 = X[y==1]

    minmax = np.min((np.max(X_0, axis=0), np.max(X_1, axis=0)), axis=0)   
    maxmin = np.max((np.min(X_0, axis=0), np.min(X_1, axis=0)), axis=0)
    maxmax = np.max((np.max(X_0, axis=0), np.max(X_1, axis=0)), axis=0)
    minmin = np.min((np.min(X_0, axis=0), np.min(X_1, axis=0)), axis=0)

    f_overlap = np.max((np.zeros((X.shape[1])), (minmax - maxmin)), axis=0)
    f_range = maxmax - minmin

    return np.nanprod(f_overlap/f_range)

def f3(X, y):
    """
    Calculates the Maximum individual feature efficiency (F3) metric.

    Measure describes the efficiency of each feature in the separation of classes. It considers the maximum value among all features.

    .. math::

        F3=\min^{m}_{i=1}{\\frac{n_o(f_i)}{n}}

    :type X: array-like, shape (n_samples, n_features)
    :param X: Dataset
    :type y: array-like, shape (n_samples)
    :param y: Labels of binary classification task ([0,1])

    :rtype: float
    :returns: F3 score
    """

    X = np.copy(X)
    y = np.copy(y)

    X_0 = X[y==0]
    X_1 = X[y==1]

    minmax = np.min((np.max(X_0, axis=0), np.max(X_1, axis=0)), axis=0)   
    maxmin = np.max((np.min(X_0, axis=0), np.min(X_1, axis=0)), axis=0)

    n = np.zeros((X.shape[1]))

    for i in range(X.shape[1]):
        # Modifiaction to original equation: replace '>' and '<' with '>=' and '<=' to return a value of 1 when all instances overlap
        n[i] = np.sum((X[:,i]>=maxmin[i]) & (X[:,i]<=minmax[i]))
    
    return np.min(n/X.shape[0])


def f4(X, y):
    """
    Calculates the Collective feature efficiency (F4) metric.

    The measure describes an overview of how the features work together. The instances separated by the most discriminant feature that was not used already are excluded from the further analysis. The process continues until all instances are classified or all features are used. The measure is calculated according to the number of instances in the overlapping region after all features were used and the total number of samples.

    .. math::

        F4=\\frac{n_o(f_{min}(T_l))}{n}

    :type X: array-like, shape (n_samples, n_features)
    :param X: Dataset
    :type y: array-like, shape (n_samples)
    :param y: Labels of binary classification task ([0,1])

    :rtype: float
    :returns: F4 score
    """

    X = np.copy(X)
    y = np.copy(y)

    T = np.zeros(X.shape[0]) #classified = 1, else 0 
    features_used = np.zeros(X.shape[1]) #used = 1, else 0

    while np.sum(T)<X.shape[0] and np.sum(features_used)<X.shape[1]:
        m1 = T==0
        m2 = y==0
        m3 = y==1

        X_0_temp = X[m1 & m2]
        X_1_temp = X[m1 & m3]
        # not yet classified
        if X_0_temp.shape[0]==0 or X_1_temp.shape[0]==0:
            break

        minmax_temp = np.min((np.max(X_0_temp, axis=0), np.max(X_1_temp, axis=0)), axis=0)   
        maxmin_temp = np.max((np.min(X_0_temp, axis=0), np.min(X_1_temp, axis=0)), axis=0)
      
        n_mask = np.zeros((X.shape)).astype(bool)
        n_mask[T==1,:] = 1

        for i in range(X.shape[1]):
            a = X[:,i]<maxmin_temp[i]
            b = X[:,i]>minmax_temp[i] 
            # not overlaping

            n_mask[:,i] = a | b

        discriminance = np.sum(n_mask, axis=0)
        most_discriminant = np.flip(np.argsort(discriminance))

        for i in most_discriminant:
            if features_used[i]==0:
                features_used[i] = 1
                T[n_mask[:,i]] = 1
                break

    return np.sum(T==0)/y.shape[0]


from . import classification
from . import regression
from .classification.feature_based import f1, f1v, f2, f3, f4
from .classification.linearity import l1, l2, l3
from .classification.neighborhood import n1, n2, n3, n4, t1, lsc
from .classification.network import density, clsCoef, hubs
from .classification.dimensionality import  t2, t3, t4
from .classification.class_imbalance import c1, c2
from .ComplexityCalculator import ComplexityCalculator
from ._version import __version__

__all__ = [
    "ComplexityCalculator",
    "__version__",
    "f1", 
    "f1v", 
    "f2",
    "f3",
    "f4",
    "l1",
    "l2",
    "l3",
    "n1",
    "n2",
    "n3",
    "n4",
    "t1",
    "lsc",
    "density",
    "clsCoef",
    "hubs",
    "t2",
    "t3",
    "t4",
    "c1",
    "c2"
]



from . import classification as c
from . import regression as r
import numpy as np
import warnings

C_METRICS = [c.f1, c.f1v, c.f2, c.f3, c.f4, c.l1, c.l2, c.l3, c.n1, c.n2, c.n3, c.n4,
             c.t1, c.lsc, c.density, c.clsCoef, c.hubs, c.t2, c.t3, c.t4, c.c1, c.c2]
C_COLORS = ['#FD0100', '#F76915', '#EEDE04', '#A0D636', '#2FA236', '#333ED4']
C_RANGES = {'FB': 5, 'LR': 3, 'NB': 6,
            'NE': 3, 'DM': 3, 'CI': 2}
C_WEIGHTS = np.ones((22))

R_METRICS = [r.c1, r.c2, r.c3, r.c4, r.l1, r.l2, r.s1, r.s2, r.s3, r.l3, r.s4, r.t2]
R_COLORS = ['#FD0100', '#F76915', '#EEDE04', '#A0D636']
R_RANGES = {'FC': 4, 'LR': 2, 'SM': 3, 'GT': 3} 
R_WEIGHTS = np.ones((12))
R_WEIGHTS[[0,1,-1]]=-1

MULTICLASS_STRATEGIES = ['ova', 'ovo'] 
MODES = ['classification', 'regression']

class ComplexityCalculator:
    """
    Complexity Calculator Class.
    
    A class that allows to determine all or selected metrics for a given data set. The report can be returned both as a simple vector of metrics, as well as a dictionary containing all set parameters and visualization in the form of a radar.

    :type metrics: list, optional (default=all the metrics avalable in problexity)
    :param metrics: List of classification complexity measures used to validate a given set.
    :type mode: string, optional (default=classification)
    :param mode: Recognition task for which metrics should be calculated. Might be selected between `classification` and `regression`.
    :type multiclass_strategy: string, optional (default=ova)
    :param multiclass_strategy: Strategy used for multiclass metric integration. Might be selected between `ova` and `ovo`.
    :type ranges: dict, optional (default=all the default six groups of metrics)
    :param ranges: Configuration of radar visualisation, allowing to group metrics by color.
    :type colors: list, optional (default=six-color palette)
    :param colors: List of colors assigned to groups on radar visualisation.
    :type weights: list, optional (default=list of weights, where weight are equal to 1 for all measures where simpler problems have smaller value, otherwise -1)
    :param weights: List of weights taken into account in score() procedure.

    :vartype complexity: list
    :var complexity: The list of all the scores acquired with metrics defined by metrics list.
    :vartype n_samples: int
    :var n_samples: The number of samples in the fitted dataset.
    :vartype n_features: int
    :var n_features: The number of features of the fitted dataset.
    :vartype n_classes: int
    :var n_classes: The number of classes in the fitted dataset.
    :vartype classes: array-like, shape (n_classes, )
    :var classes: The class labels.
    :vartype prior_probability: array-like, shape (n_classes, )
    :var prior_probability: The prior probability of classes.

    :Examples:

    >>> from problexity import ComplexityCalculator
    >>> from sklearn.datasets import make_classification
    >>> X, y = make_classification()
    >>> cc = ComplexityCalculator().fit(X, y)
    >>> print(cc.complexity)
    [0.3158144010174404, 0.1508882806154997, 0.005974480517635054, 0.57, 0.0, 
     0.10518058962953956, 0.1, 0.07, 0.135, 0.48305940839428635, 0.27, 0.11, 
     1.0, 0.9642, 0.9892929292929293, 0.9321428571428572, 0.9297111755529109, 
     0.2, 0.16, 0.8, 0.0, 0.0]
    >>> report = cc.report()
    >>> print(report)
    {
        'n_samples': 100, 'n_features': 20, 'n_classes': 2, 
        'classes': array([0, 1]), 
        'prior_probability': array([0.5, 0.5]), 
        'score': 0.377, 
        'complexities': 
        {
            'f1': 0.316, 'f1v': 0.151, 'f2': 0.006, 'f3': 0.57, 'f4': 0.0, 
            'l1': 0.105, 'l2': 0.1, 'l3': 0.07, 
            'n1': 0.135, 'n2': 0.483, 'n3': 0.27, 'n4': 0.11, 't1': 1.0, 'lsc': 0.964, 
            'density': 0.989, 'clsCoef': 0.932, 'hubs': 0.93, 
            't2': 0.2, 't3': 0.16, 't4': 0.8, 'c1': 0.0, 'c2': 0.0
        }
    }
    """
    def __init__(self, metrics=None, colors=None, ranges=None, weights=None, mode='classification', multiclass_strategy='ovo'):
        # Initlialize configuration
        if mode not in MODES:
            raise Exception('Unsupported mode %s, must be from list %s.' % (
                mode,
                ', '.join(MODES)
            ))
        else:
            self.mode = mode
        
        # Set default metrics, colors and ranges based on mode, if not provided
        _passed_params = np.sum([_ is None for _ in [metrics, colors, ranges, weights]])
        if  _passed_params > 0 and _passed_params < 4:
            warnings.warn('Not passing all parameters in [metrics, colors, ranges, weights] will result in using default values.')
            
        if None not in [metrics, colors, ranges, weights]:
            self.metrics = metrics
            self.colors = colors
            self.ranges = ranges
            self.weights = weights
        else:
            if self.mode == 'regression':
                self.metrics = R_METRICS
                self.colors = R_COLORS
                self.ranges = R_RANGES
                self.weights = R_WEIGHTS
            else:
                self.metrics = C_METRICS
                self.colors = C_COLORS
                self.ranges = C_RANGES
                self.weights = C_WEIGHTS
        
        # Validate test configuration
        rsum = np.sum([self.ranges[k] for k in self.ranges])
        if len(self.ranges) != len(self.colors):
            raise Exception('Number of ranges and colors does not match.')
        if rsum != len(self.metrics):
            raise Exception('Ranges does not sum with number of metrics.')
        if multiclass_strategy not in MULTICLASS_STRATEGIES:
            raise Exception('Unsupported multiclass_strategy %s, must be from list %s.' % (
                multiclass_strategy,
                ', '.join(MULTICLASS_STRATEGIES)
            ))
        else:
            self.multiclass_strategy = multiclass_strategy
        
        # Get all the metric names
        self.cnames = [m.__name__ for m in self.metrics]   
        
    def fit(self, X, y):
        """
        Calculates metrics for given dataset.

        :type X: array-like, shape (n_samples, n_features)
        :param X: The input samples.
        :type y: array-like, shape (n_samples, )
        :param y: The input labels. In case of classification may represent binary or multi-class problem.

        :rtype: ComplexityCalculator class object
        :returns: ComplexityCalculator class object.
        """

        # Check is fit had been called
        # Do basic calculations
        self.n_samples, self.n_features = X.shape
        if self.mode == 'classification':
            self.classes, self.prior_probability = np.unique(y, return_counts=True)
            self.prior_probability = self.prior_probability / self.n_samples
            self.n_classes = len(self.classes)
        
        # Calculate complexity
        # Binary case
        if (self.mode == 'regression') or (self.n_classes == 2):
            self.complexity = [m(X, y) for m in self.metrics]
        else:
            # Prepare container for complexities
            complexities = []
            
            
            if self.multiclass_strategy == 'ova':
                # OVA
                for i in self.classes:
                    complexities.append([m(X, (y == i).astype(int)) for m in self.metrics])
            elif self.multiclass_strategy == 'ovo':
                # OVO
                for i in self.classes:
                    for j in self.classes:
                        if i!= j:
                            mask = (y==i) + (y==j)
                            _y = (y[mask] == j).astype(int)
                            complexities.append([m(X[mask], _y) for m in self.metrics])
              
            # Calculate mean complexity                  
            self.complexity = np.mean(np.array(complexities), axis=0).tolist()
            
            # Establish complexity deviation
            self.complexity_std = np.std(np.array(complexities), axis=0).tolist()            
            
        return self
    
    def _metrics(self):
        return [metric.__name__ for metric in self.metrics]
    
    def report(self, precision=3):
        """Returns report of problem complexity

        :type precision: int, optional (default=3)
        :param precision: The rounding precision.

        :rtype: dict
        :returns: Dictionary with complexity report
        
        :Examples:

        >>> from problexity import ComplexityCalculator
        >>> from sklearn.datasets import make_classification
        >>> X, y = make_classification()
        >>> cc = ComplexityCalculator().fit(X, y)
        >>> report = cc.report()
        >>> print(report)
        {
            'n_samples': 100, 'n_features': 20, 'n_classes': 2, 
            'classes': array([0, 1]), 
            'prior_probability': array([0.5, 0.5]), 
            'score': 0.377, 
            'complexities': 
            {
                'f1': 0.316, 'f1v': 0.151, 'f2': 0.006, 'f3': 0.57, 'f4': 0.0, 
                'l1': 0.105, 'l2': 0.1, 'l3': 0.07, 
                'n1': 0.135, 'n2': 0.483, 'n3': 0.27, 'n4': 0.11, 't1': 1.0, 'lsc': 0.964, 
                'density': 0.989, 'clsCoef': 0.932, 'hubs': 0.93, 
                't2': 0.2, 't3': 0.16, 't4': 0.8, 'c1': 0.0, 'c2': 0.0
            }
        }
        """
        self._check_is_fitted()
        report = {
            'n_samples': self.n_samples,
            'n_features': self.n_features,
            'score': np.around(self.score(), precision),
            'complexities': {}
        }
        if self.mode == 'classification':
            report.update({'n_classes': self.n_classes,
                           'classes': self.n_classes,
                           'prior_probability': self.prior_probability})

            if self.n_classes > 2:
                report.update({'complexities_std': {}})

        if (self.mode == 'regression') or (self.n_classes == 2):
            for metric, score in zip(self.metrics, self.complexity):
                report['complexities'].update({
                    metric.__name__: np.around(score,precision)
                })
        else:
            for metric, score, std in zip(self.metrics, self.complexity, self.complexity_std):
                report['complexities'].update({
                    metric.__name__: np.around(score,precision)
                })
                report['complexities_std'].update({
                    metric.__name__: np.around(std,precision)
                })

        return report

    def score(self):
        """Returns integrated score of problem complexity

        :type weights: array-like, optional (default=None), shape (n_metrics) 
        :param weights: Optional weights of metrics.

        :rtype: float
        :returns: Single score for integrated metrics
        """
        self._check_is_fitted()  

        # Check length of weights vector
        if len(self.weights) != len(self.metrics):
            raise Exception('Mismatch between number of metrics and number of weights.')

        # Normalize weights
        self.weights = np.array(self.weights) / np.sum(self.weights)
        
        # Calculate weighted score
        return np.sum(self.weights * self.complexity)
       
    def plot(self, figure, spec=(1,1,1)):
        """Returns integrated score of problem complexity

        :type weights: matplotlib figure object 
        :param weights: Figure to draw radar on.
        :type spec: tuple, optional (default=(1,1,1)) 
        :param spec: Matplotlib subplot location.
        
        :rtype: object
        :returns: Matplotlib axis object.

        .. image:: plots/radar.png

        """
        # Establish location
        ax = figure.add_subplot(*spec, projection='polar') 
        
        # Prepare calculation values
        count = 0
        index = 0
        aid = 0
        
        # Group angles
        g_angles = np.linspace(0, np.pi*2, len(self.ranges)+1)

        # Major and minor ticks
        b_ticks = np.linspace(0,1,5)
        s_ticks = np.linspace(0,1,9)
        
        # Select scale
        scale = 1.
        
        # Iterate groups
        for ridx, rname in enumerate(self.ranges):
            l = self.ranges[rname]
            c = self.colors[ridx]
            
            v = self.complexity[count:count+l]
            
            # Iterate cv up
            count += l
            index += (l-1)
            
            # Iterate leading angle
            g_a, g_b = g_angles[ridx], g_angles[ridx+1]
            g_r = np.linspace(g_a, g_b, l+1)
                        
            # Iterate metrics
            for vidx, val in enumerate(v):
                # Clip value to range
                cval = np.clip(val, 0, 1)
                grv = g_r[vidx:vidx+2]
                
                # Plot metric region
                ax.fill_between(grv*scale, [-.0,-.0], [cval,cval],
                                color=self.colors[ridx],
                                alpha=1, lw=0)
                
                # Plot baseline
                ax.plot([g_r[vidx]*scale,
                        g_r[vidx]*scale], 
                        [0,1], 
                        c='white')
                
                # Iterate major y ticks
                for tick in b_ticks:
                    ax.plot(grv*scale, [tick, tick], 
                            color='white',
                            lw=1)
                    
                # Iterate minor y ticks
                for tick in s_ticks:
                    ax.plot(grv*scale, [tick, tick], 
                            color='white',
                            lw=.5)
                        
                # Leading metric lines                
                ax.plot(np.ones(b_ticks.shape)*np.mean(grv*scale),
                        b_ticks,
                        color=c,
                        zorder=1, 
                        alpha=.25)
                
                # Pinhead
                ax.scatter([np.mean(grv*scale)],[1], color=c, zorder=10)
                
                # Value label
                ax.text(np.mean(grv*scale),1.2,
                    '%s\n%.2f' % (self.cnames[aid], val),
                    color=c,
                    ha='center',
                    va='center')
                
                # Iterate metric pointer
                aid += 1

        # Plot cleanup
        ax.spines['polar'].set_visible(False)
        ax.spines['start'].set_visible(False)
        ax.spines['end'].set_visible(False)
        ax.spines['inner'].set_visible(False)
        ax.grid(lw=0)
        ax.set_xticks([])
        ax.set_yticks([])
        
        ax.set_ylim(-.15,1.25)
        ax.text(-.15, -.15, ('%.1f' if self.mode=='classification' else '%.2f') % (self.score()*(100 if self.mode=='classification' else 1)),
                ha='center',
                va='center',
                fontsize=12,
                color='#333')
        
        return ax
    
    def _check_is_fitted(self):
        if not hasattr(self, 'complexity'):
            raise Exception('Measures not calculated, please call fit() first.')



from .correlation import c1, c2, c3, c4
from .linearity import l1, l2
from .smoothness import s1, s2, s3
from .geometry import l3, s4, t2

__all__ = [
    "c1", 
    "c2", 
    "c3",
    "c4",
    "l1",
    "l2",
    "s1",
    "s2",
    "s3",
    "l3",
    "s4",
    "t2"
]



"""
Linearity measures
"""
import numpy as np
from sklearn.svm import LinearSVC

def l1(X, y):
    """
    Calculates the Sum of the error distance by linear programming (L1) metric.

    Uses Linear SVM classifier. Measures distance of incorrectly classified samples from the SVM hyperplane.

    .. math::

        L1=\\frac{SumErrorDist}{1+SumErrorDist}

    :type X: array-like, shape (n_samples, n_features)
    :param X: Dataset
    :type y: array-like, shape (n_samples)
    :param y: Labels of binary classification task ([0,1])

    :rtype: float
    :returns: L1 score
    """    
    X = np.copy(X)
    y = np.copy(y)

    svm = LinearSVC().fit(X, y)
    incorrect = svm.predict(X)!=y
    
    w_norm = np.linalg.norm(svm.coef_)
    dist = np.abs(svm.decision_function(X) / w_norm)
    
    sed = np.sum(dist[incorrect])/y.shape[0]
    return sed/(1+sed)

def l2(X, y):
    """
    Calculates the Error rate of linear classifier (L2) metric.

    Returns error rate of Linear SVM classifer used within the dataset.

    .. math::

        L2=\\frac{\sum^{n}_{i=1}I(h(x_i)\\neq y_i)}{n}

    :type X: array-like, shape (n_samples, n_features)
    :param X: Dataset
    :type y: array-like, shape (n_samples)
    :param y: Labels of binary classification task ([0,1])

    :rtype: float
    :returns: L2 score
    """

    X = np.copy(X)
    y = np.copy(y)

    svm = LinearSVC()
    pred = svm.fit(X, y).predict(X)

    return np.sum(pred!=y)/y.shape[0]


def l3(X, y):
    """
    Calculates the Non linearity of linear classifier (L3) metric.

    Linearly interpolating instances of each class generate the additional instances of the problem, which are used to calculate this measure. The class of original instances determines the label of an augmented point. The Linear SVM classifier is used to classify the synthesized points of the dataset. The number of synthetic points is equal to the original dataset size.
    
    .. math::

        L3=\\frac{1}{l}\sum^{l}_{i=1}I(h_T(x'_i) \\neq y'_i)

    :type X: array-like, shape (n_samples, n_features)
    :param X: Dataset
    :type y: array-like, shape (n_samples)
    :param y: Labels of binary classification task ([0,1])

    :rtype: float
    :returns: L3 score
    """
    
    X = np.copy(X)
    y = np.copy(y)

    try:
        pairs_0 = np.array([np.random.choice(np.argwhere(y==0).flatten(), 2, replace=False) for i in range(np.sum(y==0))])
        pairs_1 = np.array([np.random.choice(np.argwhere(y==1).flatten(), 2, replace=False) for i in range(np.sum(y==1))])
    except:
        # Protect againt single class sample
        pairs_0 = np.array([np.random.choice(np.argwhere(y==0).flatten(), 2, replace=True) for i in range(np.sum(y==0))])
        pairs_1 = np.array([np.random.choice(np.argwhere(y==1).flatten(), 2, replace=True) for i in range(np.sum(y==1))])

    pairs = np.concatenate((pairs_0, pairs_1), axis=0)
    y_new = np.concatenate((np.zeros(pairs_0.shape[0]), np.ones(pairs_1.shape[0])), axis=0)
    rand = np.random.random(pairs.shape[0]).reshape(-1,1)

    distance = X[pairs[:,0]]-X[pairs[:,1]]    
    rand_distance = rand * distance
    
    X_new = X[pairs[:,1]] + rand_distance

    svm = LinearSVC()
    pred = svm.fit(X, y).predict(X_new)

    return np.sum(pred!=y_new)/y_new.shape[0]
